/*
 * Copyright(c) 2003 - 2005 Timothy I. Mattox.  e-mail: tmattox<at>gmail.com
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/*
 * fnn_router.c
 * generate routing table for a single node in a FNN cluster
 *
 * For more information, see http://aggregate.org/FNN/
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "fnn_router.h"

#define PORT_BIT	10
#define MAX_ETH_PORTS	8
#define MAX_SWITCH_PORTS	(1 << PORT_BIT)

typedef struct port_struct {
	unsigned short ID;/* ID of switch or node that port is connected to */
	unsigned short fspn;	/* 3 bits flags, 3 bits speed, 10 (PORT_BIT) bits portNum */
} portT;

typedef struct swnode_struct {
	portT *ports;
	unsigned short portCt, portMax;
} swnodeT;

/* first[0] == switches, first[1] == nodes */
typedef struct group_struct {
	unsigned short first[2], last[2], max[2], speed;
} grpT;

unsigned int *FNN_table = NULL;
swnodeT *snA = NULL; /* all the switches and nodes */
portT *portsA = NULL; /* all the switch ports and eths */
int portsTot = 0, portsNext = 0;

/* return 1 if successful, 0 otherwise */
int
readFullNetwork(FILE *fp, unsigned short *pcs, unsigned short *gateway)
{
	register int i, z;
	char netIDstr[20];
	char netMSKstr[20];
	char gwStr[9];
	char buf[128], *bP = NULL;
	unsigned short g, new_g;
	unsigned short IDx, IDy, new_ID;
	unsigned short speed, new_port;
	portT aPort, *portP = NULL;
	swnodeT *xP, *yP;
	unsigned int gCt = 0;
	grpT gA[26]; /* all the groups */
	/* switch #0 is a virtual black hole, nodes start at 0x8000 */
	unsigned short snCt[2] = { 1, 0x8000 };

	memset(&(buf[0]), 0, sizeof(char) * 128);

/* Example input file, followed by a crude grammar.
   The file format does NOT currently support the comments!

3 10.8.0.0 255.248.0.0 C0	# groups IP_NETWORK IP_NETMASK GatewayNode
A 128 3 17 24 2	# letter maxnodes maxeths maxswitches maxports defaultSpeed
B 64 4 9 32 2
C 16 6 2 50 2
A:
0< 0a 10a 15b 30c A32a2 A42g 57a 62a 0C4 B1a3
#(Switch 0 in group C 10GigE link, Node 1 in group B eth0 GigE)
B:
5> 0 * 1 3 7 # node 5 in group B is connected to switches 0 1 3 7

{grpCount} {IP_NETWORK} {IP_NETMASK} {gatewayEntry}
{grpLetter} {maxNodes} {maxEths} {maxSwitches} {maxPorts} {portSpeed}
...
{grpLetter} {maxNodes} {maxEths} {maxSwitches} {maxPorts} {portSpeed}
{grpLetter}:
{switchNumber}< ({nodeEntry} | {switchEntry})*
{nodeNumber}> ({nodeEntry} | {switchEntry} | '*')*

gatewayEntry:	{grpLetter}{nodeNumber}
switchEntry:	{switchNumber}{grpLetter}[portSpeed]
nodeEntry:	[grpLetter]{nodeNumber}{ethLetter}[portSpeed]
grpLetter:	{A-Z (a single uppercase letter)}
nodeNumber:	{0-65534 decimal number}
ethLetter:	{a-g (a single lowercase letter)}
portSpeed:	{0-7 log10(Mb/s), 1 == 10 Mb/s, 3 == 1 Gb/s}

*/
	if (4 != fscanf(fp,"%u %16s %16s %8s\n",
		&gCt, &(netIDstr[0]), &(netMSKstr[0]), &(gwStr[0]))) {
		fprintf(stderr, "FNN.net header line was malformed.\n");
		goto outq;
	}
//	fprintf(stderr, "%u groups, NET=%s, MASK=%s, Gateway=%s\n", gCt, netIDstr, netMSKstr, gwStr);
	if (gCt > 26) {
		fprintf(stderr, "fnn_router only supports 26 or fewer groups.\n");
		goto outq;
	}

//	if ((gA = malloc(sizeof(grpT) * gCt)) == NULL) goto out;
	memset(&(gA[0]), 0, sizeof(gA));

	for (i = 0; i < gCt; ++i) {
		unsigned short g_nCt, g_sCt;
		if (1 != fscanf(fp,"%1s", &(buf[0]))) {
			fprintf(stderr, "Group header label '%c' missing or read error.\n", i + 'A');
			goto outq;
		}
		g = buf[0] - 'A';
		if (g >= gCt) {
			fprintf(stderr, "Group header label '%c' out of range.\n", buf[0]);
			goto out;
		}
		gA[g].first[0] = snCt[0];
		gA[g].first[1] = snCt[1];
		if (5 != fscanf(fp,"%hu %hu %hu %hu %hu",
				&(g_nCt),
				&(gA[g].max[1]),
				&(g_sCt),
				&(gA[g].max[0]),
				&(gA[g].speed))) {
			fprintf(stderr, "Group header line read error.\n");
			goto out;
		}
		if (gA[g].speed > 7) gA[g].speed = 7;
		if (gA[g].max[1] > MAX_ETH_PORTS) {
			fprintf(stderr, "Group '%c', Too many ports per PE (%d > %d).\n", g + 'A', gA[g].max[1], MAX_ETH_PORTS);
			goto outq;
		}
		if (gA[g].max[0] > MAX_SWITCH_PORTS) {
			fprintf(stderr, "Group '%c', Too many ports per switch (%d > %d).\n", g + 'A', gA[g].max[0], MAX_SWITCH_PORTS);
			goto outq;
		}
		portsTot += (g_sCt * gA[g].max[0]) + (g_nCt * gA[g].max[1]);
		snCt[0] += g_sCt;
		snCt[1] += g_nCt;
		gA[g].last[0] = snCt[0] - 1;
		gA[g].last[1] = snCt[1] - 1;
		if (gA[g].last[0] >= gA[0].first[1]) {
			fprintf(stderr, "Too many switches.\n");
			goto out;
		}
		if (gA[g].last[1] < gA[g].first[1]) {
			fprintf(stderr, "Too many nodes.\n");
			goto out;
		}
	}
	snA = calloc(0x10000, sizeof(swnodeT));
	if (!snA) {
		fprintf(stderr, "calloc failed for the switch/node array.\n");
		goto outq;
	}
	portsA = calloc(portsTot, sizeof(portT));
	if (!portsA) {
		fprintf(stderr, "calloc failed for the ports array.\n");
		goto outq;
	}

	for (g = 0; g < gCt; ++g) {
		register int j;
		for (j = 0; j < 2; ++j) {
			for (i = gA[g].first[j]; i <= gA[g].last[j]; ++i) {
				snA[i].portCt = 0;
				snA[i].ports = &(portsA[portsNext]);
				snA[i].portMax = gA[g].max[j];
				portsNext += gA[g].max[j];
			}
		}
	}
	if (portsNext != portsTot) {
		fprintf(stderr, "Internal error, ports assigned != ports allocated.\n");
		goto outq;
	}

	g = 0;
	IDx = IDy = 0; /* switch #0 is a black hole */
	xP = &(snA[0]);

	while (fscanf(fp, "%127s", bP = &(buf[0])) == 1) {
		new_g = isupper(*bP) ? *(bP++) - 'A' : g;
		if (new_g >= gCt) {
			fprintf(stderr, "Group '%c' out of range.\n", new_g + 'A');
			goto out;
		}
		if (isdigit(*bP)) {
			new_ID = strtol(bP, &bP, 10);
		} else if (*bP == ':') {
			g = new_g;
			continue;
		} else if (*bP == '*') { /* skip a port/eth */
			++(xP->portCt);
			continue;
		} else {
			fprintf(stderr, "Unexpected character '%c'.\n", *bP);
			goto out;
		}
		z = *(bP++);
		speed = ((z != 0) && isdigit(*bP)) ? *bP - '0' : gA[g].speed;
		if (speed > 7) speed = 7;
		if (islower(z)) { /* it's as a node */
			IDy = new_ID + gA[new_g].first[1];
			if ((IDy > gA[new_g].last[1]) || (IDy < gA[new_g].first[1])) {
				fprintf(stderr, "Node %d of group '%c' out of range.\n", new_ID, new_g + 'A');
				goto out;
			}
			yP = &(snA[IDy]);

			new_port = z - 'a';
		} else if (z == '>') { /* start of a node's eth list */
			IDx = new_ID + gA[new_g].first[1];
			if ((IDx > gA[new_g].last[1]) || (IDx < gA[new_g].first[1])) {
				fprintf(stderr, "Node %d of group '%c' out of range.\n", new_ID, new_g + 'A');
				goto out;
			}
			xP = &(snA[IDx]);
			continue;
		} else if (z == '<') { /* start of a switch's port list */
			IDx = new_ID + gA[new_g].first[0];
			if ((IDx > gA[new_g].last[0]) || (IDx < gA[new_g].first[0])) {
				fprintf(stderr, "Switch %d of group '%c' out of range.\n", new_ID, new_g + 'A');
				goto out;
			}
			xP = &(snA[IDx]);
			continue;
		} else { /* it's as a switch */
			new_g = isupper(z) ? z - 'A' : g;
			if (new_g >= gCt) {
				fprintf(stderr, "Group '%c' out of range.\n", new_g + 'A');
				goto out;
			}

			IDy = new_ID + gA[new_g].first[0];
			if ((IDy > gA[new_g].last[0]) || (IDy < gA[new_g].first[0])) {
				fprintf(stderr, "Switch %d of group '%c' out of range.\n", new_ID, new_g + 'A');
				goto out;
			}
			yP = &(snA[IDy]);

			new_port = (yP->portCt)++;
		}

		if (new_port >= yP->portMax) {
			fprintf(stderr, "Destination port %d out of range (0-%d).\n", new_port, yP->portMax - 1);
			goto out;
		}
		if (xP->portCt >= xP->portMax) {
			fprintf(stderr, "Source port %d out of range (0-%d).\n", xP->portCt, xP->portMax - 1);
			goto out;
		}
		/* make new connection */
		aPort.ID = IDx;
		aPort.fspn = (speed<<PORT_BIT) | xP->portCt;
		portP = &(yP->ports[new_port]);
		if (portP->ID && ((portP->ID != aPort.ID) || (portP->fspn != aPort.fspn))) {
			fprintf(stderr, "Destination port %d already in use.\n", new_port);
			goto out;
		}
		*portP = aPort;

		aPort.ID = IDy;
		aPort.fspn = (speed<<PORT_BIT) | new_port;
		portP = &(xP->ports[xP->portCt]);
		if (portP->ID && ((portP->ID != aPort.ID) || (portP->fspn != aPort.fspn))) {
			fprintf(stderr, "Source port %d already in use.\n", xP->portCt);
			goto out;
		}
		*portP = aPort;

		(xP->portCt)++;
	}
	bP = &(gwStr[0]);
	g = isupper(*bP) ? *(bP++) - 'A' : 0;
	new_ID = isdigit(*bP) ? strtol(bP, &bP, 10) : 0;
	IDx = new_ID + gA[g].first[1];
	if ((IDx > gA[g].last[1]) || (IDx < gA[g].first[1])) {
		fprintf(stderr, "Gateway is out of range (node %d of group '%c').\n", new_ID, g + 'A');
		goto outq;
	}
	*gateway = IDx & ~0x8000;
	*pcs = snCt[1] & ~0x8000;

	return 1;
out:
	buf[127] = 0; /* NULL terminate the buffer just in case */
	fprintf(stderr, "Token in parser buffer was '%s'\n", buf);
outq:
	return 0;
}


/* fill in FNN route table for a node in a multi-FNN cluster
 *
 * NOTE: this version does NOT use the inter-group information!
 * It assumes that all NICs can be used to communicate between groups,
 * as if all switches from each group were connected to a common
 * top-level switch.  This is a BAD assumption, but does the "right thing"
 * for the FNN clusters in current existence. -- Tim 20050713
 * Proper handling of inter-group connections is planned in a future release.
 * Proper handling of SFNN thru-node routing is planned in a future release.
 */

void
gen_FNN_routes(unsigned short pcs, unsigned short whichPC)
{
	register unsigned short i, first, last;
	first = 0x8000;
	last = first + pcs - 1;
	whichPC += first;

	FNN_table = calloc(pcs, sizeof(unsigned int));
	if (FNN_table == NULL) return;

	for (i = first; i <= last; ++i) {
		register unsigned int entry = 0;
		register int a, b, e, ct = 0;
		if (i == whichPC) continue;
		for (b = 0; b < snA[i].portMax; ++b) {
			/* clear flag bits */
			snA[i].ports[b].fspn &= (1 << (PORT_BIT + 3)) - 1;
		}

		for (a = 0; a < snA[whichPC].portMax; ++a) {
			register int a_s = (snA[whichPC].ports[a].fspn>>PORT_BIT)&0x7;
			e = snA[whichPC].ports[a].ID;
			if (e == 0) continue; /* this NIC is not in the FNN */
			if (e == i) { /* direct connection found */
				b = snA[whichPC].ports[a].fspn & 0x7;
fprintf(stderr, "directConnect %d %d %d\n", i - first, a, b);
				entry = (b << 3) | a;
				ct = 1;
				goto processEntry;
			}
			for (b = 0; b < snA[i].portMax; ++b) {
				if ((e == snA[i].ports[b].ID) & !(snA[i].ports[b].fspn & 0x2000)) {
					register int b_s = (snA[i].ports[b].fspn>>PORT_BIT)&0x7;
					if (b_s <= a_s) snA[i].ports[b].fspn |= 0x2000; /* mark as used */
					entry |= ((b << 3) | a) << (ct++ * 6);
					if (ct == 5) goto processEntry;
					if (a_s <= b_s) break; /* don't reuse port 'a' */
				}
			}
		}
/* start of temporary hack to avoid doing thru routing */
		if (ct > 0) goto processEntry;
/* match up eth#'s in increasing order if all are the same speed */
/* otherwise, make each slower nick connect to a single faster nick */
		a = 0; b = 0;
		while ((a < snA[whichPC].portMax) && (b < snA[i].portMax)) {
			register int a_s, b_s;
			if (snA[whichPC].ports[a].ID == 0) {
				++a;
				continue;
			}
			if (snA[i].ports[b].ID == 0) {
				++b;
				continue;
			}
			entry |= ((b << 3) | a) << (ct++ * 6);
/* make faster link's sticky... re-use them */
/* this assumes all the links of an individual node are the same speed */
			a_s = (snA[whichPC].ports[a].fspn>>PORT_BIT)&0x7;
			b_s = (snA[i].ports[b].fspn>>PORT_BIT)&0x7;
//fprintf(stderr, "%d %d %d %d ", a, b, a_s, b_s);
			if (a_s <= b_s) ++a;
			if (b_s <= a_s) ++b;
//fprintf(stderr, "(%d %d)\n", a, b);

			if (ct == 5) goto processEntry;
		}
/* end of temporary hack to avoid doing thru routing */

processEntry:
		switch (ct) {
		case 0:	/* through routing */
			entry = 0 /* nextNodeID(whichPC, i) << 14 */;
			ct = 2;
			break;
		case 1:	entry |= (entry << 12) | (entry << 6);
			entry |= (entry << 18); /* drop this line and make ct = 3 if desired */
			ct = 5;
			break;
		case 2:	entry |= (entry << 12);
			ct = 4;
			break;
		default: /* nothing to do ... */
			break;
		}
		entry = (entry << 2) | (ct - 2);
		
		FNN_table[i - first] = entry;
// fprintf(stderr, "%03d: %010o %d\n", i - first, entry >> 2, (entry & 0x3) ? (entry & 0x3) + 2 : 0);
	}
/* Do real thru routing here... fill in routes for distance two nodes */
}

/* save the FNN route table for a node in a FNN cluster */
void
save_FNN_routes(unsigned short pcs, unsigned short whichPC, unsigned short gateway)
{
	int len, sendCt, packSize = 1024 /* getpagesize() */ / sizeof(unsigned int);
	unsigned short nodesSent;
	int res;
	unsigned int *FNNrtp;

	fprintf(stderr, "Saving routes for PE%05hu of %hu with gateway as PE%05hu\n",
		whichPC, pcs, gateway);

	FNNrtp = malloc(packSize * sizeof(unsigned int));
	if ((FNNrtp == NULL) && (FNN_table == NULL)) return;

	FNNrtp[0] = (FNNRTP_FreeAllRts << 28);
	FNNrtp[1] = (FNNRTP_ClearStats << 28);
	FNNrtp[2] = (FNNRTP_SetGateway << 28) | gateway;
	len = 3;
	nodesSent = 0;
	while (nodesSent < pcs) {
		sendCt = packSize - len - 1;
		if (sendCt > (pcs - nodesSent)) sendCt = pcs - nodesSent;
		FNNrtp[len++] = (FNNRTP_NewRoutes << 28) | (sendCt << 16) | nodesSent;
		memcpy((void *) &(FNNrtp[len]), (void *) &(FNN_table[nodesSent]), sendCt * sizeof(unsigned int));
		len += sendCt;
		nodesSent += sendCt;
		res = fwrite((void *) FNNrtp,sizeof(unsigned int),len,stdout);
		if (res != len) {
			fprintf(stderr,
				"Write to stdout of the FNN_routes table failed\n");
			break;
		}
		fflush(stdout);
//		fprintf(stderr, "nodesSent=%d len=%d\n", nodesSent, len);
		len = 0;
	}
	free(FNNrtp);
}


int
main(register int argc,
register char **argv)
{
	unsigned short pcs, whichPC, gateway;
	struct in_addr addr;
	int res = -1;

	if (argc < 2) {
		fprintf(stderr,
"Usage: %s [IPADDR, PE_Number, or hostname with trailing PE_number] < raw_FNN_pattern_data > FNN_routes\n",
			argv[0]);
		exit(1);
	}

	if (inet_aton(argv[1], &addr)) {
		whichPC = (unsigned short) (ntohl(addr.s_addr) & 0xffff);
	} else if (!(res = sscanf(argv[1], "%hu", &whichPC))) {
		res = sscanf(argv[1], "%*[^0123456789]%hu", &whichPC);
	}
	//fprintf(stderr, "res=%d, whichPC=%d\n", res, whichPC);
	fprintf(stderr, "Computing routes for PE%05hu\n", whichPC);

	if (!readFullNetwork(stdin, &pcs, &gateway)) {
		fprintf(stderr, "An error occured when reading the FNN net description, exiting.\n");
		exit(1);
	}

	gen_FNN_routes(pcs, whichPC);
	save_FNN_routes(pcs, whichPC, gateway);

	exit(0);
}
